@@ -7,7 +7,7 @@
  * is found, writes the GPS data into the EXIF data
  * in the photo. For future reference... */
 
-/* Copyright 2005 Daniel Foote.
+/* Copyright 2005-2018 Daniel Foote, Dan Fandrich.
  *
  * This file is part of gpscorrelate.
  *
@@ -35,21 +35,24 @@
 #include "correlate.h"
 #include "unixtime.h"
 
+#define MIN(a,b) (((a)<(b))?(a):(b))
+
 /* Internal functions used to make it work. */
-void Round(struct GPSPoint* First, struct GPSPoint* Result,
-		struct CorrelateOptions* Options, time_t PhotoTime);
-void Interpolate(struct GPSPoint* First, struct GPSPoint* Result,
-		struct CorrelateOptions* Options, time_t PhotoTime);
+static void Round(const struct GPSPoint* First, struct GPSPoint* Result,
+		  time_t PhotoTime);
+static void Interpolate(const struct GPSPoint* First, struct GPSPoint* Result,
+			time_t PhotoTime);
 
 /* This function returns a GPSPoint with the point selected for the
  * file. This allows us to do funky stuff like not actually write
  * the files - ie, just correlate and keep into memory... */
 
-struct GPSPoint* CorrelatePhoto(char* Filename,
+struct GPSPoint* CorrelatePhoto(const char* Filename,
 		struct CorrelateOptions* Options)
 {
 	/* Read out the timestamp from the EXIF data. */
-	char* TimeTemp; int IncludesGPS = 0;
+	char* TimeTemp;
+	int IncludesGPS = 0;
 	TimeTemp = ReadExifDate(Filename, &IncludesGPS);
 	if (!TimeTemp)
 	{
@@ -58,15 +61,41 @@ struct GPSPoint* CorrelatePhoto(char* Filename,
 		 * will appear on the console. Otherwise, we were
 		 * returned here due to the lack of exif tags. */
 		Options->Result = CORR_NOEXIFINPUT;
-		return 0;
+		return NULL;
 	}
-	if (IncludesGPS)
+	if (IncludesGPS && !Options->OverwriteExisting)
 	{
 		/* Already have GPS data in the file!
 		 * So we can't do this again... */
 		Options->Result = CORR_GPSDATAEXISTS;
-		return 0;
+		free(TimeTemp);
+		return NULL;
 	}
+	if (Options->AutoTimeZone)
+	{
+		/* Use the local time zone as of the date of first picture
+		 * as the time for correlating all the remainder. */
+		time_t RealTime;
+
+		/* PhotoTime isn't a true epoch time, but is rather out
+		 * by the local offset from UTC */
+		time_t PhotoTime =
+			ConvertToUnixTime(TimeTemp, EXIF_DATE_FORMAT, 0, 0);
+
+		/* Extract the component time values */
+		struct tm *PhotoTm = gmtime(&PhotoTime);
+
+		/* Then create a true epoch-based local time, including DST */
+		PhotoTm->tm_isdst = -1;
+		RealTime = mktime(PhotoTm);
+
+		/* Finally, RealTime is the proper Epoch time of the photo.
+		 * The difference from PhotoTime is the time zone offset. */
+		Options->TimeZoneHours = (PhotoTime - RealTime) / 3600;
+		Options->TimeZoneMins = ((PhotoTime - RealTime) % 3600) / 60;
+		Options->AutoTimeZone = 0;
+	}
+	//printf("Using offset %02d:%02d\n", Options->TimeZoneHours, Options->TimeZoneMins);
 
 	/* Now convert the time into Unixtime. */
 	time_t PhotoTime =
@@ -81,55 +110,64 @@ struct GPSPoint* CorrelatePhoto(char* Filename,
 	 * be freed for us. */
 	free(TimeTemp);
 
-	/* Check to see if MinTime and MaxTime are filled in.
-	 * If not, fill them in. */
-	if (Options->MinTime == 0 && Options->MaxTime == 0)
+	/* Search the list of GPS tracks to find one containing the range
+	 * we're interested in. Options points to an array with the last
+	 * entry denoted by a NULL Points pointer. */
+	int TrackNum;
+	for (TrackNum = 0; Options->Track[TrackNum].Points; ++TrackNum)
 	{
-		/* Alright, fill them in! 
-		 * Requires us to go through the list and keeping
-		 * the biggest and smallest. The list should,
-		 * however, be sorted. But we do it this way anyway. */
-		struct GPSPoint* Fill = NULL;
-		Options->MinTime = Options->Points->Time;
-		for (Fill = Options->Points; Fill; Fill = Fill->Next)
-		{
-			/* Ignore trackseg markers... */
-			if (Fill->Lat == 1000 && Fill->Long == 1000)
-				continue;
-			/* Check the Min time */
-			if (Fill->Time < Options->MinTime)
-				Options->MinTime = Fill->Time;
-			/* Check the Max time */
-			if (Fill->Time > Options->MaxTime) 
-				Options->MaxTime = Fill->Time;
-		}
+		/* Check that the photo is within the times that
+		 * our tracks are for. Can't really match it if
+		 * we were not logging when it was taken.
+		 * Note: photos taken between logging sessions of the
+		 * same file will still make it inside of this. In
+		 * some cases, it won't matter, but if it does, then
+		 * keep this in mind!! */
+		if ((PhotoTime >= Options->Track[TrackNum].MinTime) &&
+		    (PhotoTime <= Options->Track[TrackNum].MaxTime))
+			break;
 	}
-
-	/* Check that the photo is within the times that
-	 * our tracks are for. Can't really match it if
-	 * we were not logging when it was taken. */
-	/* Note: photos taken between logging sessions of the
-	 * same file will still make it inside of this. In
-	 * some cases, it won't matter, but if it does, then
-	 * keep this in mind!! */
-	if ((PhotoTime < Options->MinTime) ||
-			(PhotoTime > Options->MaxTime))
-	{
-		/* Outside the range. Abort. */
+	if (!Options->Track[TrackNum].Points) {
+		/* All tracks were outside the time range. Abort. */
 		Options->Result = CORR_NOMATCH;
-		return 0;
+		return NULL;
 	}
 
 	/* Time to run through the list, and see if our PhotoTime
 	 * is in between two points. Alternately, it might be
 	 * exactly on a point... even better... */
-	struct GPSPoint* Search;
-	struct GPSPoint* Actual = malloc(sizeof(struct GPSPoint));
+	const struct GPSPoint* Search;
+	struct GPSPoint* Actual = (struct GPSPoint*) malloc(sizeof(struct GPSPoint));
 
 	Options->Result = CORR_NOMATCH; /* For convenience later */
-	
-	for (Search = Options->Points; Search; Search = Search->Next)
+
+	for (Search = Options->Track[TrackNum].Points; Search; Search = Search->Next)
 	{
+		/* First test: is it exactly this point? */
+		if (PhotoTime == Search->Time)
+		{
+			/* This is the point, exactly.
+			 * Copy out the data and return that. */
+			Actual->Lat = Search->Lat;
+			Actual->LatDecimals = Search->LatDecimals;
+			Actual->Long = Search->Long;
+			Actual->LongDecimals = Search->LongDecimals;
+			Actual->Elev = Search->Elev;
+			Actual->ElevDecimals = Search->ElevDecimals;
+			Actual->Time = Search->Time;
+
+			Options->Result = CORR_OK;
+			break;
+		}
+
+		/* Sanity check / track segment fix: is the photo time before
+		 * the current point? If so, we've gone past it. Hrm. */
+		if (Search->Time > PhotoTime)
+		{
+			Options->Result = CORR_NOMATCH;
+			break;
+		}
+
 		/* Sanity check: we need to peek at the next point.
 		 * Make sure we can. */
 		if (Search->Next == NULL) break;
@@ -154,14 +192,6 @@ struct GPSPoint* CorrelatePhoto(char* Filename,
 			}
 		}
 
-		/* Sanity check / track segment fix: is the photo time before
-		 * the current point? If so, we've gone past it. Hrm. */
-		if (Search->Time > PhotoTime)
-		{
-			Options->Result = CORR_NOMATCH;
-			break;
-		}
-		
 		/* Sort of sanity check: is this photo inside our
 		 * "feather" time? If not, abort. */
 		if (Options->FeatherTime)
@@ -185,20 +215,6 @@ struct GPSPoint* CorrelatePhoto(char* Filename,
 			}
 		} /* endif (Options->Feather) */
 		
-		/* First test: is it exactly this point? */
-		if (PhotoTime == Search->Time)
-		{
-			/* This is the point, exactly.
-			 * Copy out the data and return that. */
-			Actual->Lat = Search->Lat;
-			Actual->Long = Search->Long;
-			Actual->Elev = Search->Elev;
-			Actual->Time = Search->Time;
-
-			Options->Result = CORR_OK;
-			break;
-		}
-		
 		/* Second test: is it between this and the
 		 * next point? */
 		if ((PhotoTime > Search->Time) &&
@@ -207,18 +223,16 @@ struct GPSPoint* CorrelatePhoto(char* Filename,
 			/* It is between these points.
 			 * Unless told otherwise, we interpolate.
 			 * If not interpolating, we round to nearest.
-			 * If points are eqidistant, we round down. */
+			 * If points are equidistant, we round down. */
 			if (Options->NoInterpolate)
 			{
 				/* No interpolation. Round. */
-				Round(Search, Actual,
-						Options, PhotoTime);
+				Round(Search, Actual, PhotoTime);
 				Options->Result = CORR_ROUND;
 				break;
 			} else {
 				/* Interpolate away! */
-				Interpolate(Search, Actual,
-						Options, PhotoTime);
+				Interpolate(Search, Actual, PhotoTime);
 				Options->Result = CORR_INTERPOLATED;
 				break;
 			}
@@ -256,14 +270,14 @@ struct GPSPoint* CorrelatePhoto(char* Filename,
 	 * and return nothing. */
 	free(Actual);
 	return NULL;
-};
+}
 
-void Round(struct GPSPoint* First, struct GPSPoint* Result,
-		struct CorrelateOptions* Options, time_t PhotoTime)
+void Round(const struct GPSPoint* First, struct GPSPoint* Result,
+	   time_t PhotoTime)
 {
 	/* Round the point between the two points - ie, it will end
 	 * up being one or the other point. */
-	struct GPSPoint* CopyFrom = NULL;
+	const struct GPSPoint* CopyFrom = NULL;
 
 	/* Determine the difference between the two points. 
 	 * We're using the scale function used by interpolate.
@@ -283,16 +297,19 @@ void Round(struct GPSPoint* First, struct GPSPoint* Result,
 
 	/* Copy the numbers over... */
 	Result->Lat = CopyFrom->Lat;
+	Result->LatDecimals = CopyFrom->LatDecimals;
 	Result->Long = CopyFrom->Long;
+	Result->LongDecimals = CopyFrom->LongDecimals;
 	Result->Elev = CopyFrom->Elev;
+	Result->ElevDecimals = CopyFrom->ElevDecimals;
 	Result->Time = CopyFrom->Time;
 
 	/* Done! */
 	
 }
 
-void Interpolate(struct GPSPoint* First, struct GPSPoint* Result,
-		struct CorrelateOptions* Options, time_t PhotoTime)
+void Interpolate(const struct GPSPoint* First, struct GPSPoint* Result,
+		 time_t PhotoTime)
 {
 	/* Interpolate between the two points. The first point
 	 * is First, the other First->Next. Results into Result. */
@@ -304,15 +321,19 @@ void Interpolate(struct GPSPoint* First, struct GPSPoint* Result,
 	double Scale = (double)First->Next->Time - (double)First->Time;
 	Scale = ((double)PhotoTime - (double)First->Time) / Scale;
 
-	/* Now calculate the Lattitude. */
+	/* Now calculate the Latitude. */
 	Result->Lat = First->Lat + ((First->Next->Lat - First->Lat) * Scale);
+	Result->LatDecimals = MIN(First->LatDecimals, First->Next->LatDecimals);
 
 	/* And the longitude. */
 	Result->Long = First->Long + ((First->Next->Long - First->Long) * Scale);
+	Result->LongDecimals = MIN(First->LongDecimals, First->Next->LongDecimals);
 
-	/* And the elevation. If elevation wasn't set, it should be zero.
-	 * Which works quite fine for us. */
+	/* And the elevation. If elevation wasn't set, it should be zero with
+	 * a negative ElevDecimals, which will cause it to be dropped
+	 * when written. */
 	Result->Elev = First->Elev + ((First->Next->Elev - First->Elev) * Scale);
+	Result->ElevDecimals = MIN(First->ElevDecimals, First->Next->ElevDecimals);
 
 	/* The time is not interpolated, but matches photo. */
 	Result->Time = PhotoTime;
