@@ -6,7 +6,7 @@
  * functions in correlate.c.
  */
 
-/* Copyright 2005 Daniel Foote.
+/* Copyright 2005-2018 Daniel Foote, Dan Fandrich.
  *
  * This file is part of gpscorrelate.
  *
@@ -29,10 +29,12 @@
  * Not really sure if this is needed, but... */
 struct CorrelateOptions {
 	int NoWriteExif;
+	int OverwriteExisting;
 	int NoInterpolate;
 	int NoChangeMtime;
 	int TimeZoneHours;  /* To add to photos to make them UTC. */
 	int TimeZoneMins;
+	int AutoTimeZone;
 	int FeatherTime;
 	char* Datum;     /* Datum of the data; when writing. */
 	int DoBetweenTrkSeg; /* Match between track segments. */
@@ -40,14 +42,12 @@ struct CorrelateOptions {
 	
 	int Result;
 
-	time_t MinTime;  /* Calculated on first pass. Used to   */
-	time_t MaxTime;  /* determine when to throw photos out. */
-
 	int PhotoOffset; /* Offset applied to Photo time. This is ADDED to PHOTO TIME
 			    to make it match GPS time. In seconds. 
 			    This is (GPS - Photo) */
 
-	struct GPSPoint* Points; /* Points to use... */
+	struct GPSTrack *Track; /* Pointer to array of tracks to use. The last
+				   track must be entirely zeros. */
 };
 
 /* Return codes in order:
@@ -76,6 +76,5 @@ struct CorrelateOptions {
 #define CORR_GPSDATAEXISTS  8
 
 
-struct GPSPoint* CorrelatePhoto(char* Filename, 
+struct GPSPoint* CorrelatePhoto(const char* Filename, 
 		struct CorrelateOptions* Options);
-
