@@ -6,7 +6,7 @@
  * containing GPS data.
  */
 
-/* Copyright 2005 Daniel Foote.
+/* Copyright 2005-2016 Daniel Foote, Dan Fandrich.
  *
  * This file is part of gpscorrelate.
  *
@@ -31,14 +31,17 @@
 #include <libxml/parser.h>
 #include <libxml/tree.h>
 
+#include "i18n.h"
 #include "gpx-read.h"
 #include "unixtime.h"
 #include "gpsstructure.h"
+#include "latlong.h"
 
-struct GPSPoint* FirstPoint;
-struct GPSPoint* LastPoint;
+/* Pointers to the first and last points, used during parsing */
+static struct GPSPoint* FirstPoint;
+static struct GPSPoint* LastPoint;
 
-void ExtractTrackPoints(xmlNodePtr Start)
+static void ExtractTrackPoints(xmlNodePtr Start)
 {
 	/* The pointer passed to us should be the start
 	 * of a heap of trkpt's. So walk though them,
@@ -46,15 +49,15 @@ void ExtractTrackPoints(xmlNodePtr Start)
 	xmlNodePtr Current = NULL;
 	xmlNodePtr CCurrent = NULL;
 	xmlAttrPtr Properties = NULL;
-	char* Lat;
-	char* Long;
-	char* Elev;
-	char* Time;
+	const char* Lat;
+	const char* Long;
+	const char* Elev;
+	const char* Time;
 
 	for (Current = Start; Current; Current = Current->next)
 	{
 		if ((Current->type == XML_ELEMENT_NODE) &&
-			(strcmp((char *)Current->name, "trkpt") == 0))
+			(strcmp((const char *)Current->name, "trkpt") == 0))
 		{
 			/* This is indeed a trackpoint. Extract! */
 
@@ -72,13 +75,13 @@ void ExtractTrackPoints(xmlNodePtr Start)
 					Properties;
 					Properties = Properties->next)
 			{
-				if (strcmp((char *)Properties->name, "lat") == 0)
+				if (strcmp((const char *)Properties->name, "lat") == 0)
 				{
-					Lat = (char *)Properties->children->content;
+					Lat = (const char *)Properties->children->content;
 				}
-				if (strcmp((char *)Properties->name, "lon") == 0)
+				if (strcmp((const char *)Properties->name, "lon") == 0)
 				{
-					Long = (char *)Properties->children->content;
+					Long = (const char *)Properties->children->content;
 				}
 			}
 
@@ -90,20 +93,20 @@ void ExtractTrackPoints(xmlNodePtr Start)
 					CCurrent;
 					CCurrent = CCurrent->next)
 			{
-				if (strcmp((char *)CCurrent->name, "ele") == 0)
+				if (strcmp((const char *)CCurrent->name, "ele") == 0)
 				{
 					if (CCurrent->children)
-						Elev = (char *)CCurrent->children->content;
+						Elev = (const char *)CCurrent->children->content;
 				}
-				if (strcmp((char *)CCurrent->name, "time") == 0)
+				if (strcmp((const char *)CCurrent->name, "time") == 0)
 				{
 					if (CCurrent->children)
-						Time = (char *)CCurrent->children->content;
+						Time = (const char *)CCurrent->children->content;
 				}
 			}
 
 			/* Check that we have all the data. If we're missing something,
-			 * then skip this point... */
+			 * then skip this point... NOTE: Elev is not required. */
 			if (Time == NULL || Long == NULL || Lat == NULL)
 			{
 				/* Missing some data. */
@@ -116,27 +119,34 @@ void ExtractTrackPoints(xmlNodePtr Start)
 			if (FirstPoint)
 			{
 				/* Ok, adding to the list... */
-				LastPoint->Next = malloc(sizeof(struct GPSPoint));
+				LastPoint->Next = (struct GPSPoint*) malloc(sizeof(struct GPSPoint));
 				LastPoint = LastPoint->Next;
 				LastPoint->Next = NULL;
 			} else {
 				/* This is the first one. */
-				FirstPoint = malloc(sizeof(struct GPSPoint));
+				FirstPoint = (struct GPSPoint*) malloc(sizeof(struct GPSPoint));
+				FirstPoint->Next = NULL;
 				LastPoint = FirstPoint;
 			}
 
 			/* Clear the structure first... */
 			LastPoint->Lat = 0;
+			LastPoint->LatDecimals = 0;
 			LastPoint->Long = 0;
+			LastPoint->LongDecimals = 0;
 			LastPoint->Elev = 0;
+			LastPoint->ElevDecimals = -1; // default meaning no altitude was found
 			LastPoint->Time = 0;
 			LastPoint->EndOfSegment = 0;
 
 			/* Write the data into LastPoint, which should be a new point. */
 			LastPoint->Lat = atof(Lat);
+			LastPoint->LatDecimals = NumDecimals(Lat);
 			LastPoint->Long = atof(Long);
+			LastPoint->LongDecimals = NumDecimals(Long);
 			if (Elev) {
 				LastPoint->Elev = atof(Elev);
+				LastPoint->ElevDecimals = NumDecimals(Elev);
 			}
 			LastPoint->Time = ConvertToUnixTime(Time, GPX_DATE_FORMAT, 0, 0);
 			
@@ -144,6 +154,7 @@ void ExtractTrackPoints(xmlNodePtr Start)
 			printf("TrackPoint. Lat %s (%f), Long %s (%f). Elev %s (%f), Time %d.\n",
 					Lat, atof(Lat), Long, atof(Long), Elev, atof(Elev),
 					ConvertToUnixTime(Time, GPX_DATE_FORMAT, 0, 0));
+			printf("Decimals %d %d %d\n", LastPoint->LatDecimals, LastPoint->LongDecimals, LastPoint->ElevDecimals);
 			*/
 			
 					
@@ -153,7 +164,7 @@ void ExtractTrackPoints(xmlNodePtr Start)
 	/* Return control to the recursive function... */
 }
 
-void FindTrackSeg(xmlNodePtr Start)
+static void FindTrackSeg(xmlNodePtr Start)
 {
 	/* Go recursive till we find a <trgseg> tag. */
 	xmlNodePtr Current = NULL;
@@ -161,7 +172,7 @@ void FindTrackSeg(xmlNodePtr Start)
 	for (Current = Start; Current; Current = Current->next)
 	{
 		if ((Current->type == XML_ELEMENT_NODE) &&
-			(strcmp((char *)Current->name, "trkseg") == 0))
+			(strcmp((const char *)Current->name, "trkseg") == 0))
 		{
 			/* Found it... the children should
 			 * all be trkpt's. */
@@ -180,8 +191,31 @@ void FindTrackSeg(xmlNodePtr Start)
 
 }
 
+/* Determines and stores the min and max times from the GPS track */
+static void GetTrackRange(struct GPSTrack* Track)
+{
+	if (Track->Points == NULL)
+		return;
 
-struct GPSPoint* ReadGPX(char* File)
+	/* Requires us to go through the list and keep
+	 * the biggest and smallest. The list should,
+	 * however, be sorted. But we do it this way anyway. */
+	const struct GPSPoint* Fill = NULL;
+	Track->MaxTime = 0;
+	Track->MinTime = Track->Points->Time;
+	for (Fill = Track->Points; Fill; Fill = Fill->Next)
+	{
+		/* Check the Min time */
+		if (Fill->Time < Track->MinTime)
+			Track->MinTime = Fill->Time;
+		/* Check the Max time */
+		if (Fill->Time > Track->MaxTime) 
+			Track->MaxTime = Fill->Time;
+	}
+}
+
+
+int ReadGPX(const char* File, struct GPSTrack* Track)
 {
 	/* Init the libxml library. Also checks version. */
 	LIBXML_TEST_VERSION
@@ -190,10 +224,10 @@ struct GPSPoint* ReadGPX(char* File)
 	
 	/* Read the GPX data from file. */
 	GPXData = xmlParseFile(File);
-	
 	if (GPXData == NULL)
 	{
-		fprintf(stderr, "Failed to parse GPX data from %s.\n", File);
+		fprintf(stderr, _("Failed to parse GPX data from %s.\n"), File);
+		xmlCleanupParser();
 		return 0;
 	}
 
@@ -203,18 +237,22 @@ struct GPSPoint* ReadGPX(char* File)
 
 	if (GPXRoot == NULL)
 	{
-		fprintf(stderr, "GPX file has no root. Not healthy.\n");
+		fprintf(stderr, _("Invalid GPX file has no root.\n"));
+		xmlFreeDoc(GPXData);
+		xmlCleanupParser();
 		return 0;
 	}
 
 	/* Check that this is indeed a GPX - the root node
 	 * should be "gpx". */
-	if (strcmp((char *)GPXRoot->name, "gpx") == 0)
+	if (strcmp((const char *)GPXRoot->name, "gpx") == 0)
 	{
 		/* Ok, it is a GPX file. */
 	} else {
 		/* Not valid. */
-		fprintf(stderr, "Invalid GPX file.\n");
+		fprintf(stderr, _("Invalid GPX file.\n"));
+		xmlFreeDoc(GPXData);
+		xmlCleanupParser();
 		return 0;
 	}
 
@@ -238,7 +276,7 @@ struct GPSPoint* ReadGPX(char* File)
 	 * puts in for me... imagine having to write an XML parser!
 	 * Nasty.) */
 	/* Before we go into this function, we also setlocale to "C".
-	 * The GPX def indicates that the decimal seperator should be
+	 * The GPX def indicates that the decimal separator should be
 	 * ".", but certain locales specify otherwise. Which has caused issues.
 	 * So we set the locale for this function, and then revert it.
 	 */
@@ -246,8 +284,7 @@ struct GPSPoint* ReadGPX(char* File)
 	FirstPoint = NULL;
 	LastPoint = NULL;
 	
-	char* OldLocale = setlocale(LC_NUMERIC, NULL);
-	setlocale(LC_NUMERIC, "C");
+	char* OldLocale = setlocale(LC_NUMERIC, "C");
 	
 	FindTrackSeg(GPXRoot);
 
@@ -257,16 +294,21 @@ struct GPSPoint* ReadGPX(char* File)
 	xmlFreeDoc(GPXData);
 	xmlCleanupParser();
 
-	return FirstPoint;
-};
+	Track->Points = FirstPoint;
+
+	/* Find the time range for this track */
+	GetTrackRange(Track);
+
+	return 1;
+}
 
 
-void FreePointList(struct GPSPoint* List)
+void FreeTrack(struct GPSTrack* Track)
 {
 	/* Free the memory associated with the
 	 * GPSPoint list... */
 	struct GPSPoint* NextFree = NULL;
-	struct GPSPoint* CurrentFree = List;
+	struct GPSPoint* CurrentFree = Track->Points;
 	while (1)
 	{
 		if (CurrentFree == NULL) break;
@@ -274,5 +316,5 @@ void FreePointList(struct GPSPoint* List)
 		free(CurrentFree);
 		CurrentFree = NextFree;
 	}
-};
-		
+	Track->Points = NULL;
+}
