@@ -6,7 +6,7 @@
  * in gpx-read.c.
  */
 
-/* Copyright 2005 Daniel Foote.
+/* Copyright 2005-2012 Daniel Foote, Dan Fandrich.
  *
  * This file is part of gpscorrelate.
  *
@@ -25,5 +25,7 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
-struct GPSPoint* ReadGPX(char* File);
-void FreePointList(struct GPSPoint* List);
+struct GPSTrack;
+
+int ReadGPX(const char* File, struct GPSTrack* Track);
+void FreeTrack(struct GPSTrack* Track);
