@@ -7,7 +7,7 @@
  * Writes the output back into the GPS exif tags.
  */
 
-/* Copyright 2005 Daniel Foote.
+/* Copyright 2005-2019 Daniel Foote, Dan Fandrich.
  *
  * This file is part of gpscorrelate.
  *
@@ -31,26 +31,30 @@
 #include <time.h>
 #include <getopt.h>
 #include <string.h>
+#include <locale.h>
+#include <math.h>
 
-#ifndef WIN32
-  #include <termios.h>
-#endif
-
+#include "i18n.h"
 #include "gpsstructure.h"
 #include "exif-gps.h"
 #include "unixtime.h"
 #include "gpx-read.h"
+#include "latlong.h"
 #include "correlate.h"
 
+#define GPS_EXIT_WARNING 2
+
 /* Command line options structure. */
-static struct option program_options[] = {
+static const struct option program_options[] = {
 	{ "gps", required_argument, 0, 'g' },
+	{ "latlong", required_argument, 0, 'l' },
 	{ "timeadd", required_argument, 0, 'z'},
 	{ "no-interpolation", no_argument, 0, 'i'},
 	{ "help", no_argument, 0, 'h'},
 	{ "verbose", no_argument, 0, 'v'},
 	{ "datum", required_argument, 0, 'd'},
 	{ "no-write", no_argument, 0, 'n'},
+	{ "replace", no_argument, 0, 'R'},
 	{ "max-dist", required_argument, 0, 'm'},
 	{ "show", no_argument, 0, 's'},
 	{ "machine", no_argument, 0, 'o'},
@@ -65,47 +69,71 @@ static struct option program_options[] = {
 };
 
 /* Function to print the version - near the top for easy modification. */
-void PrintVersion(char* ProgramName)
+static void PrintVersion(const char* ProgramName)
 {
-	printf("%s, v 1.6.1, 13 February 2010. Daniel Foote, 2005-2010. GNU GPL.\n", ProgramName);
+	printf(_("%s, ver. %s. %s\n"),
+	       ProgramName, PACKAGE_VERSION,
+	       _("Copyright (C) 2005-2019 Daniel Foote, Dan Fandrich"));
 }
 
 /* Function to print the usage info. */
-void PrintUsage(char* ProgramName)
+static void PrintUsage(const char* ProgramName)
 {
-	printf("Usage: %s --gps|-g file.gpx [options] file1.jpg file2.jpg ...\n", ProgramName);
-	printf("--gps or -g file.gpx: required, specifies GPX file with GPS data.\n");
-	printf("--timeadd or -z +/-XX[:XX]: time to add to GPS data to make it match photos.\n");
-	printf("   GPS data is in UTC; photos are not likely to be in UTC. Enter the\n");
-	printf("   timezone used when taking the photos: eg, +8 for Perth.\n");
-	printf("--no-interpolation or -i: disable interpolation between points.\n");
-	printf("   Interpolation is linear, points rounded if disabled.\n");
-	printf("--verbose or -v: show what has been selected.\n");
-	printf("--datum or -d datum: specify measurement datum. If not set, WGS-84 used.\n");
-	printf("--no-write or -n: do not write the exif data. Useful with --verbose.\n");
-	printf("--max-dist or -m time: max time outside points that photo will be matched.\n");
-	printf("   Time is in seconds.\n");
-	printf("--show or -s: Just show the GPS data from the given files, if it exists.\n");
-	printf("--machine or -o: Just show the GPS data from the given files, machine readable output.\n");
-	printf("--remove or -r: Strip GPS tags from the given files, and then quit.\n");
-	printf("--ignore-tracksegs or -t: Interpolate between track segments too.\n");
-	printf("--no-mtime or -M: Don't change mtime of modified files.\n");
-	printf("--fix-datestamps or -f: Fix broken GPS datestamps written with versions < 1.5.2\n");
-	printf("--degmins or -p: Write location as DD MM.MM as was default before < 1.5.3.\n");
-	printf("--photooffset or -O <seconds>: Offset added to photo time to make it match the GPS.\n");
-	printf("--help or -h: display usage/help message.\n");
-	printf("--version or -V: display version information.\n");
+	printf(_("Usage: %s [options] file.jpg ...\n"), ProgramName);
+	puts(  _("-g, --gps file.gpx       Specifies GPX file with GPS data"));
+	puts(  _("-l, --latlong LAT,LONG[,E] Specifies latitude/longitude/elevation directly"));
+	puts(  _("-z, --timeadd +/-HH[:MM] Time to add to GPS data to make it match photos"));
+	puts(  _("-i, --no-interpolation   Disable interpolation between points; interpolation\n"
+	         "                         is linear, points rounded if disabled"));
+	puts(  _("-d, --datum DATUM        Specify measurement datum (defaults to WGS-84)"));
+	puts(  _("-n, --no-write           Do not write the EXIF data. Useful with --verbose"));
+	puts(  _("-R, --replace            Overwrite any GPS tags already in the image file"));
+	puts(  _("-m, --max-dist SECS      Max time outside points that photo will be matched"));
+	puts(  _("-s, --show               Just show the GPS data from the given files"));
+	puts(  _("-o, --machine            Similar to --show but with machine-readable output"));
+	puts(  _("-r, --remove             Strip GPS tags from the given files"));
+	puts(  _("-t, --ignore-tracksegs   Interpolate between track segments, too"));
+	puts(  _("-M, --no-mtime           Don't change mtime of modified files"));
+	puts(  _("-f, --fix-datestamps     Fix broken GPS datestamps written with ver. < 1.5.2"));
+	puts(  _("    --degmins            Write location as DD MM.MM (was default before v1.5.3)"));
+	puts(  _("-O, --photooffset SECS   Offset added to photo time to make it match the GPS"));
+	puts(  _("-h, --help               Display this help message"));
+	puts(  _("-v, --verbose            Show more detailed output"));
+	puts(  _("-V, --version            Display version information"));
+}
+
+/* CSV escape a string by doubling quotation marks.
+ * A pointer to a malloced string is returned which must be freed by the caller.
+ */
+static char *CsvEscape(const char *str)
+{
+	char *newstr = (char*)malloc(strlen(str)*2+1); // worst-case size
+	char *cpy = newstr;
+	if (!newstr)
+		return newstr;
+	do {
+		*cpy = *str;
+		if (*cpy++ == '"')
+			*cpy++ = '"';
+	} while (*str++);
+	return newstr;
 }
 
 /* Display the information from an existing file. */
-void ShowFileDetails(char* File, int MachineReadable)
+static int ShowFileDetails(const char* File, int MachineReadable)
 {
 	double Lat, Long, Elev;
 	int IncludesGPS = 0;
-	char* Time = NULL;
-	Lat = Long = Elev = 0;
+	Lat = Long = 0;
+	Elev = NAN; /* Elevation is optional, so this means it's missing */
+	char* Time = ReadExifData(File, &Lat, &Long, &Elev, &IncludesGPS);
+	int rc = 1;
+	char* OldLocale = NULL;
 
-	Time = ReadExifData(File, &Lat, &Long, &Elev, &IncludesGPS);
+	if (MachineReadable)
+	{
+		OldLocale = setlocale(LC_NUMERIC, "C");
+	}
 
 	if (Time)
 	{
@@ -115,48 +143,72 @@ void ShowFileDetails(char* File, int MachineReadable)
 			 * it machine readable. */
 			if (MachineReadable)
 			{
-				printf("\"%s\",\"%s\",%f,%f,%f\n",
-					File, Time, Lat, Long, Elev);
+				char *EscapedFile = CsvEscape(File);
+				if (!EscapedFile) {
+					fprintf(stderr, _("Out of memory.\n"));
+					exit(EXIT_FAILURE);
+				}
+				printf("\"%s\",\"%s\",%f,%f,",
+					EscapedFile, Time, Lat, Long);
+
+				if (!isnan(Elev))
+					printf("%.3f", Elev);
+				printf("\n");
+				free(EscapedFile);
 			} else {
-				printf("%s: %s, Lat %f, Long %f, Elevation %f.\n",
-					File, Time, Lat, Long, Elev);
+				printf(_("%s: %s, Lat %f, Long %f, Elevation "),
+					File, Time, Lat, Long);
+				if (!isnan(Elev))
+					printf("%.3f", Elev);
+                                else
+					printf(_("(unknown)"));
+
+				printf(".\n");
 			}
 		} else {
 			/* Don't display anything if we want machine
 			 * readable data and there is no data. */
 			if (!MachineReadable)
 			{
-				printf("%s: %s, No GPS Data.\n",
+				printf(_("%s: %s, No GPS Data.\n"),
 					File, Time);
 			}
 		}
 	} else {
-		/* Say that there was no data, except if we want
+		/* Say that there was no data & return error, except if we want
 		 * machine readable output */
 		if (!MachineReadable)
 		{
-			printf("%s: No EXIF data.\n", File);
+			printf(_("%s: No EXIF data.\n"), File);
+			rc = 0;
 		}
 	}
 
 	free(Time);
+
+	if (MachineReadable) {
+		setlocale(LC_NUMERIC, OldLocale);
+	}
+
+	return rc;
 }
 			
 /* Remove all GPS exif tags from a file. Not really that useful, but... */
-/* BUG: presently removes tags even on read-only files. Hmm. */
-void RemoveGPSTags(char* File, int NoChangeMtime)
+static int RemoveGPSTags(const char* File, int NoChangeMtime, int NoWriteExif)
 {
-	if (RemoveGPSExif(File, NoChangeMtime))
+	if (RemoveGPSExif(File, NoChangeMtime, NoWriteExif))
 	{
-		printf("%s: Removed GPS tags.\n", File);
+		printf(_("%s: Removed GPS tags.\n"), File);
+		return 1;
 	} else {
-		printf("%s: Tag removal failure.\n", File);
+		printf(_("%s: Tag removal failure.\n"), File);
+		return 0;
 	}
 }
 
 /* Fix GPSDatestamp tags, if they were incorrect, as found with versions
  * earlier than 1.5.2. */
-void FixDatestamp(char* File, int AdjustmentHours, int AdjustmentMinutes, int NoWriteExif)
+static int FixDatestamp(const char* File, int AdjustmentHours, int AdjustmentMinutes, int NoWriteExif)
 {
 	/* Read the timestamp data. */
 	char DateStamp[12];
@@ -164,22 +216,23 @@ void FixDatestamp(char* File, int AdjustmentHours, int AdjustmentMinutes, int No
 	char CombinedTime[24];
 	int IncludesGPS = 0;
 	char* OriginalDateStamp = NULL;
+	int rc = 1;
 
 	OriginalDateStamp = ReadGPSTimestamp(File, DateStamp, TimeStamp, &IncludesGPS);
 
 	if (OriginalDateStamp == NULL)
 	{
-		printf("%s: No EXIF data.\n", File);
+		printf(_("%s: No EXIF data.\n"), File);
+		rc = 0;
 	} else if (IncludesGPS == 0) {
-		printf("%s: No GPS data.\n", File);
+		printf(_("%s: No GPS data.\n"), File);
+		rc = 0;
 	} else {
 		/* Check the timestamp. */
 		time_t PhotoTime = ConvertToUnixTime(OriginalDateStamp, EXIF_DATE_FORMAT,
 			AdjustmentHours, AdjustmentMinutes);
 		
-		strcpy(CombinedTime, DateStamp);
-		strcat(CombinedTime, " ");
-		strcat(CombinedTime, TimeStamp);
+		snprintf(CombinedTime, sizeof(CombinedTime), "%s %s", DateStamp, TimeStamp);
 
 		time_t GPSTime = ConvertToUnixTime(CombinedTime, EXIF_DATE_FORMAT, 0, 0);
 
@@ -189,35 +242,40 @@ void FixDatestamp(char* File, int AdjustmentHours, int AdjustmentMinutes, int No
 			 * GPSTimestamp, which was wrong too. */
 			if (!NoWriteExif)
 			{
-				WriteFixedDatestamp(File, PhotoTime);
+				rc = WriteFixedDatestamp(File, PhotoTime);
 			}
 			char PhotoTimeFormat[100];
 			char GPSTimeFormat[100];
-			char *tmp = ctime(&PhotoTime);
-			strncpy(PhotoTimeFormat, tmp, 100);
-			tmp = ctime(&GPSTime);
-			strncpy(GPSTimeFormat, tmp, 100);
-			printf("%s: Wrong timestamp:\n   Photo:     %s   GPS:       %s   Corrected: %s", 
+
+			strftime(PhotoTimeFormat, sizeof(PhotoTimeFormat),
+				 "%a %b %d %H:%M:%S %Y UTC", gmtime(&PhotoTime));
+			strftime(GPSTimeFormat, sizeof(GPSTimeFormat),
+				 "%a %b %d %H:%M:%S %Y UTC", gmtime(&GPSTime));
+			printf(_("%s: Wrong timestamp:\n   Photo:     %s\n"
+				 "   GPS:       %s\n   Corrected: %s\n"),
 					File, PhotoTimeFormat, GPSTimeFormat, PhotoTimeFormat);
 		} else {
 			/* Inside the range. Do nothing! */
-			printf("%s: Timestamp is OK: Photo %s (localtime), GPS %s (UTC).\n", File,
-					OriginalDateStamp, CombinedTime);
+			printf(_("%s: Timestamp is OK: Photo %s (localtime), GPS %s (UTC).\n"),
+			       File, OriginalDateStamp, CombinedTime);
 		}
 	}
 
 	free(OriginalDateStamp);
+	return rc;
 }
 
 int main(int argc, char** argv)
 {
-	/* Say hello. */
-	printf("EXIF-GPS Photo matching program.\n");
-	printf("Daniel Foote, 2005.\n\n");
+	/* Initialize locale & gettext */
+	setlocale (LC_ALL, "");
+	(void) bindtextdomain(TEXTDOMAIN, PACKAGE_LOCALE_DIR);
+	(void) textdomain(TEXTDOMAIN);
 
 	/* If you didn't pass any arguments... */
 	if (argc == 1)
 	{
+		PrintVersion(argv[0]);
 		PrintUsage(argv[0]);
 		exit(EXIT_SUCCESS);
 	}
@@ -226,13 +284,17 @@ int main(int argc, char** argv)
 	/* But first, some variables to store stuff in. */
 	int c;
 	
-	char* GPSData = NULL;        /* Filename of the file with the GPS data. */
-	char* TimeAdjustment = NULL; /* Time adjustment, as passed to program. */
+	struct GPSTrack* Track = NULL;/* Array of lists of GPS waypoints. The
+					 final entry of all 0 signals the end. */
+	int NumTracks = 0;	     /* Number of track structures at Track,
+					not including the terminating entry. */
+	int HaveTimeAdjustment = 0;  /* Whether -z option was given. */
 	int TimeZoneHours = 0;       /* Integer version of the timezone. */
 	int TimeZoneMins = 0;
 	char* Datum = NULL;          /* Datum of input GPS data. */
 	int Interpolate = 1;         /* Do we interpolate? By default, yes. */
 	int NoWriteExif = 0;         /* Do we not write to file? By default, no. */
+	int OverwriteExisting = 0;   /* Do we overwrite existing tags? By default, no */
 	int ShowDetails = 0;         /* Do we show lots of details? By default, no. */
 	int FeatherTime = 0;         /* The "feather" time, in seconds. 0 = disabled. */
 	int ShowOnlyDetails = 0;
@@ -243,12 +305,21 @@ int main(int argc, char** argv)
 	int FixDatestamps = 0;
 	int DegMinSecs = 1;
 	int PhotoOffset = 0;
+	struct GPSPoint LatLong;
+
+	/* Create the empty terminating array entry */
+	Track = (struct GPSTrack*) calloc(1, sizeof(*Track));
+	if (!Track)
+	{
+		fprintf(stderr, _("Out of memory.\n"));
+		exit(EXIT_FAILURE);
+	}
 
 	while (1)
 	{
 		/* Call getopt to do all the hard work
 		 * for us... */
-		c = getopt_long(argc, argv, "g:z:ihvd:m:nsortMVfO:",
+		c = getopt_long(argc, argv, "g:z:il:hvd:m:nsortRMVfO:",
 				program_options, 0);
 
 		if (c == -1) break;
@@ -258,13 +329,58 @@ int main(int argc, char** argv)
 		{
 			case 'g':
 				/* This parameter specifies the GPS data.
-				 * It must be present. */
+				 * It or 'l' must be present at least once. */
 				if (optarg)
 				{
-					GPSData = malloc((sizeof(char) * strlen(optarg)) + 1);
-					strncpy(GPSData, optarg, strlen(optarg)+1);
+					int HaveTrack;
+
+					/* Read the XML file into memory and extract the "points". */
+					printf(_("Reading GPS Data..."));
+					fflush(stdout);
+					HaveTrack = ReadGPX(optarg, &Track[NumTracks]);
+					printf("\n");
+					if (!HaveTrack)
+					{
+						exit(EXIT_FAILURE);
+					}
+
+					/* Make room for a new end-of-array entry */
+					++NumTracks;
+					Track = (struct GPSTrack*) realloc(Track, sizeof(*Track)*(NumTracks+1));
+					if (!Track)
+					{
+						fprintf(stderr, _("Out of memory.\n"));
+						exit(EXIT_FAILURE);
+					}
+					memset(&Track[NumTracks], 0, sizeof(*Track));
 				}
 				break;
+			case 'l':
+				/* This parameter specifies a direct latitude/longitude
+				   coordinate to use for all images.
+				   It or 'g' must be present at least once. */
+				if (!ParseLatLong(optarg, &LatLong))
+				{
+					printf(_("Error parsing location.\n"));
+					exit(EXIT_FAILURE);
+				}
+				if (!MakeTrackFromLatLong(&LatLong, &Track[NumTracks]))
+				{
+					printf(_("Out of memory\n"));
+					exit(EXIT_FAILURE);
+				}
+
+				/* Make room for a new end-of-array entry */
+				++NumTracks;
+				Track = (struct GPSTrack*) realloc(Track, sizeof(*Track)*(NumTracks+1));
+				if (!Track)
+				{
+					printf(_("Out of memory\n"));
+					exit(EXIT_FAILURE);
+				}
+				memset(&Track[NumTracks], 0, sizeof(*Track));
+				break;
+
 			case 'z':
 				/* This parameter specifies the time to add to the
 				 * GPS data to make it match the timezone for
@@ -272,12 +388,25 @@ int main(int argc, char** argv)
 				/* We only store it here, convert it to numbers later. */
 				if (optarg)
 				{
-					TimeAdjustment = malloc((sizeof(char) * strlen(optarg)) + 1);
-					strncpy(TimeAdjustment, optarg, strlen(optarg)+1);
+					/* Break up the adjustment and convert to numbers. */
+					if (strstr(optarg, ":"))
+					{
+						/* Found colon. Split into two. */
+						sscanf(optarg, "%d:%d", &TimeZoneHours, &TimeZoneMins);
+						if (TimeZoneHours < 0)
+						    TimeZoneMins *= -1;
+					} else {
+						/* No colon. Just parse. */
+						TimeZoneHours = atoi(optarg);
+					}
+					HaveTimeAdjustment = 1;
 				}
 				break;
 			case 'O':
-				PhotoOffset = atoi(optarg);
+				if (optarg)
+				{
+					PhotoOffset = atoi(optarg);
+				}
 				break;
 			case 'i':
 				/* This option disables interpolation. */
@@ -285,27 +414,30 @@ int main(int argc, char** argv)
 				break;
 			case 'v':
 				/* This option asks us to show more info. */
-#ifdef WIN32
-				printf("--verbose does not work on win32\n");
-#else
+				PrintVersion(argv[0]);
 				ShowDetails = 1;
-#endif
 				break;
 			case 'd':
 				/* This option specifies a Datum, if other than WGS-84. */
 				if (optarg)
 				{
-					Datum = malloc((sizeof(char) * strlen(optarg)) + 1);
-					strncpy(Datum, optarg, strlen(optarg)+1);
+					Datum = strdup(optarg);
 				}
 				break;
 			case 'n':
 				/* This option specifies not to write to file. */
 				NoWriteExif = 1;
 				break;
+			case 'R':
+				/* This option specifies to overwrite existing GPS tags. */
+				OverwriteExisting = 1;
+				break;
 			case 'm':
 				/* This option gives us the allowable "feather" time. */
-				FeatherTime = atoi(optarg);
+				if (optarg)
+				{
+					FeatherTime = atoi(optarg);
+				}
 				break;
 			case 'h':
 				/* Display the help/usage information. And then quit. */
@@ -315,6 +447,8 @@ int main(int argc, char** argv)
 			case 'V':
 				/* Display version information, and then quit. */
 				PrintVersion(argv[0]);
+				printf(_("This is free software; see the source for copying conditions.  There is NO\n"
+					 "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"));
 				exit(EXIT_SUCCESS);
 				break;
 			case 'f':
@@ -347,8 +481,7 @@ int main(int argc, char** argv)
 				break;
 			case '?':
 				/* Unrecognised option. Or, missing argument. */
-				/* We should inform the user and let them correct this. */
-				printf("Next time, please pass a parameter with that!\n");
+				/* The user has already been informed, so just exit. */
 				exit(EXIT_FAILURE);
 				break;
 			default:
@@ -358,35 +491,37 @@ int main(int argc, char** argv)
 		} /* End switch(c) */
 	} /* End While(1) */
 	
-	/* Check to see if the user passed some files work with. Not much
+	/* Check to see if the user passed some files to work with. Not much
 	 * good if they didn't. */
 	if (optind < argc)
 	{
 		/* You passed some files. Handy! */
 	} else {
 		/* Hmm. It seems there were no other files... that doesn't work. */
-		printf("Nice try! However, next time, pass a few JPEG files to match!\n");
-		exit(EXIT_SUCCESS);
+		fprintf(stderr, _("At least one image file name must be given.\n"));
+		exit(EXIT_FAILURE);
 	}
 
 	/* If we only wanted to display info on the passed photos, do so now. */
 	if (ShowOnlyDetails)
 	{
+		int result = 1;
 		while (optind < argc)
 		{
-			ShowFileDetails(argv[optind++], MachineReadable);
+			result = ShowFileDetails(argv[optind++], MachineReadable) && result;
 		}
-		exit(EXIT_SUCCESS);
+		exit(result ? EXIT_SUCCESS : EXIT_FAILURE);
 	}
 
 	/* If we wanted to delete tags, do this now. */
 	if (RemoveTags)
 	{
+		int result = 1;
 		while (optind < argc)
 		{
-			RemoveGPSTags(argv[optind++], NoChangeMtime);
+			result = RemoveGPSTags(argv[optind++], NoChangeMtime, NoWriteExif) && result;
 		}
-		exit(EXIT_SUCCESS);
+		exit(result ? EXIT_SUCCESS : EXIT_FAILURE);
 	}
 
 	/* If we wanted to fix datestamps, do this now.
@@ -394,78 +529,33 @@ int main(int argc, char** argv)
 	 * earlier than 1.5.2. */
 	if (FixDatestamps)
 	{
-		if (TimeAdjustment == NULL)
+		if (!HaveTimeAdjustment)
 		{
-			printf("You must give a time adjustment for the photos with -z to fix photos.\n");
-			exit(EXIT_SUCCESS);
+			fprintf(stderr, _("A time offset must be given with the -z option to fix photos.\n"));
+			exit(EXIT_FAILURE);
 		}
 		
-		/* Break up the adjustment and convert to numbers. */
-		if (strstr(TimeAdjustment, ":"))
-		{
-			/* Found colon. Split into two. */
-			sscanf(TimeAdjustment, "%d:%d", &TimeZoneHours, &TimeZoneMins);
-			if (TimeZoneHours < 0)
-			    TimeZoneMins *= -1;
-		} else {
-			/* No colon. Just parse. */
-			TimeZoneHours = atoi(TimeAdjustment);
-		}
-
+		int result = 1;
 		while (optind < argc)
 		{
-			FixDatestamp(argv[optind++], TimeZoneHours, TimeZoneMins, NoWriteExif);
+			result = FixDatestamp(argv[optind++], TimeZoneHours, TimeZoneMins, NoWriteExif) && result;
 		}
-		exit(EXIT_SUCCESS);
+		exit(result ? EXIT_SUCCESS : EXIT_FAILURE);
 	}
 
 	/* Set up any other command line options... */
 	if (!Datum)
 	{
-		Datum = malloc((sizeof(char) * strlen("WGS-84")) + 1);
-		strcpy(Datum, "WGS-84");
-	}
-	if (TimeAdjustment)
-	{
-		/* Break up the adjustment and convert to numbers. */
-		if (strstr(TimeAdjustment, ":"))
-		{
-			/* Found colon. Split into two. */
-			sscanf(TimeAdjustment, "%d:%d", &TimeZoneHours, &TimeZoneMins);
-			if (TimeZoneHours < 0)
-			    TimeZoneMins *= -1;
-		} else {
-			/* No colon. Just parse. */
-			TimeZoneHours = atoi(TimeAdjustment);
-		}
-
-		/* printf("Time zone: %d : %d.\n", TimeZoneHours, TimeZoneMins); */
-			
+		Datum = strdup("WGS-84");
 	}
 
-	/* See if the user did pass any GPS data. 
-	 * If not, don't continue. */
-	if (GPSData == NULL)
+	if (!NumTracks)
 	{
-		printf("You need to pass some GPS data! Otherwise, nothing to match from!\n");
-		exit(EXIT_FAILURE);
-	}
-	
-	/* Read the XML file into memory and extract the "points". */
-	/* The returned pointer is the start of a singly-linked list. */
-	printf("Reading GPS Data...");
-	struct GPSPoint* Points = ReadGPX(GPSData);
-	printf("\n");
-
-	if (Points)
-	{
-		/* GPS Data was read correctly. */
-	} else {
 		/* GPS Data was not read correctly... */
 		/* Tell the user we are bailing.
 		 * Not really required, seeing as ReadGPX should
 		 * inform the user anyway... but, doesn't hurt! */
-		printf("Failure reading/processing GPS data.\n");
+		fprintf(stderr, _("Cannot continue since no GPS data is available.\n"));
 		exit(EXIT_FAILURE);
 	}
 
@@ -473,14 +563,16 @@ int main(int argc, char** argv)
 	 * If we're not being verbose. Otherwise, this would be pointless. */
 	if (!ShowDetails)
 	{
-		printf("Legend: . = Ok, / = Interpolated, < = Rounded, - = No match, ^ = Too far.\n");
-		printf("        w = Write Fail, ? = No exif date, ! = GPS already present.\n");
+		printf(_("Legend: . = Ok, / = Interpolated, < = Rounded, - = No match, ^ = Too far\n"
+			 "        w = Write Fail, ? = No EXIF date, ! = GPS already present\n"));
 	}
 
 	/* Set up our options structure for the correlation function. */
 	struct CorrelateOptions Options;
 	Options.NoWriteExif   = NoWriteExif;
+	Options.OverwriteExisting = OverwriteExisting;
 	Options.NoInterpolate = (Interpolate ? 0 : 1);
+	Options.AutoTimeZone  = !HaveTimeAdjustment;
 	Options.TimeZoneHours = TimeZoneHours;
 	Options.TimeZoneMins  = TimeZoneMins;
 	Options.FeatherTime   = FeatherTime;
@@ -490,38 +582,16 @@ int main(int argc, char** argv)
 	Options.DegMinSecs    = DegMinSecs;
 	Options.PhotoOffset   = PhotoOffset;
 
-	Options.MinTime = 0;
-	Options.MaxTime = 0;
-
-	Options.Points = Points;
-
-	/* Twig with the terminal settings to make the single character
-	 * output stuff work right. */
-#ifndef WIN32
-	struct termios initial_settings, new_settings;
-	if (!ShowDetails)
-	{
-		tcgetattr(fileno(stdout),&initial_settings);
-		new_settings = initial_settings;
-		new_settings.c_lflag &= ~ICANON;
-		new_settings.c_cc[VMIN] = 1;
-		new_settings.c_cc[VTIME] = 0;
-		new_settings.c_lflag &= ~ISIG;
-		if(tcsetattr(fileno(stdout), TCSANOW, &new_settings) != 0) {
-			/* Oops. Oh well. Didn't work. */
-			printf("Debug: Broken tty set.\n");
-		}
-	}
-#endif
+	Options.Track         = Track;
 
 	/* Make it all look nice and pretty... so the user knows what's going on. */
-	printf("\nCorrelate: ");
+	printf(_("\nCorrelate: "));
 	if (ShowDetails) printf("\n");
 	
 	/* A few variables that we'll require later. */
 	struct GPSPoint* Result;
 	char* File;
-	/* Including stats on what happenned. */
+	/* Including stats on what happened. */
 	int MatchExact = 0;
 	int MatchInter = 0;
 	int MatchRound = 0;
@@ -551,7 +621,7 @@ int main(int argc, char** argv)
 				MatchExact++;
 				if (ShowDetails)
 				{
-					printf("%s: Exact match: ", File);
+					printf(_("%s: Exact match: "), File);
 				} else {
 					printf(".");
 				}
@@ -561,7 +631,7 @@ int main(int argc, char** argv)
 				MatchInter++;
 				if (ShowDetails)
 				{
-					printf("%s: Interpolated: ", File);
+					printf(_("%s: Interpolated: "), File);
 				} else {
 					printf("/");
 				}
@@ -571,7 +641,7 @@ int main(int argc, char** argv)
 				MatchRound++;
 				if (ShowDetails)
 				{
-					printf("%s: Rounded: ", File);
+					printf(_("%s: Rounded: "), File);
 				} else {
 					printf("<");
 				}
@@ -581,7 +651,7 @@ int main(int argc, char** argv)
 				WriteFail++;
 				if (ShowDetails)
 				{
-					printf("%s: Exif write failure: ", File);
+					printf(_("%s: EXIF write failure: "), File);
 				} else {
 					printf("w");
 				}
@@ -589,10 +659,14 @@ int main(int argc, char** argv)
 			if (ShowDetails)
 			{
 				/* Print out the "point". */
-				printf("Lat %f, Long %f, Elev %f.\n",
-					Result->Lat, Result->Long,
-					Result->Elev);
+				printf(_("Lat %f, Long %f, Elev "),
+					Result->Lat, Result->Long);
+				if (Result->ElevDecimals >=0)
+					printf("%.3f.\n", Result->Elev);
+				else
+					printf(_("(unknown).\n"));
 			}
+			free(Result);
 			/* Ok, that's all from this part... */
 		} else {
 			/* We got nothing back. One of a few errors. */
@@ -601,7 +675,7 @@ int main(int argc, char** argv)
 				NotMatched++;
 				if (ShowDetails)
 				{
-					printf("%s: No match.\n", File);
+					printf(_("%s: No match.\n"), File);
 				} else {
 					printf("-");
 				}
@@ -611,7 +685,7 @@ int main(int argc, char** argv)
 				TooFar++;
 				if (ShowDetails)
 				{
-					printf("%s: Too far from nearest point.\n", File);
+					printf(_("%s: Too far from nearest point.\n"), File);
 				} else {
 					printf("^");
 				}
@@ -621,7 +695,7 @@ int main(int argc, char** argv)
 				NoDate++;
 				if (ShowDetails)
 				{
-					printf("%s: No date exif tag present.\n", File);
+					printf(_("%s: No EXIF date tag present.\n"), File);
 				} else {
 					printf("?");
 				}
@@ -631,7 +705,7 @@ int main(int argc, char** argv)
 				GPSPresent++;
 				if (ShowDetails)
 				{
-					printf("%s: GPS Data already present.\n", File);
+					printf(_("%s: GPS Data already present.\n"), File);
 				} else {
 					printf("!");
 				}
@@ -639,6 +713,9 @@ int main(int argc, char** argv)
 			/* Handled all those errors, now... */
 		} /* End if Result. */
 
+		/* Display the character code immediately */
+		fflush(stdout);
+
 		/* And, once we've got here, we've finished with that file.
 		 * We can now do the next one. Now wasn't that too easy? */
 		
@@ -647,32 +724,45 @@ int main(int argc, char** argv)
 	/* Right, so now we're done. That really wasn't that hard. Right? */
 
 	/* Add a new line if we were doing the not-show-details thing. */
-#ifndef WIN32
 	if (!ShowDetails)
 	{
 		printf("\n");
-		tcsetattr(fileno(stdout), TCSANOW, &initial_settings);
 	}
-#endif
 
-	/* Print details of what happenned. */
-	printf("\nCompleted correlation process.\n");
-	printf("Matched: %5d (%d Exact, %d Interpolated, %d Rounded).\n",
+	/* Print details of what happened. */
+	printf(_("\nCompleted correlation process.\n"));
+	if (ShowDetails)
+		/* This has to be shown at the end in case auto time zone
+		 * was used, since it isn't known before the first file
+		 * is processed. */
+		printf(_("Used time zone offset %d:%02d\n"),
+		       Options.TimeZoneHours, abs(Options.TimeZoneMins));
+	printf(_("Matched: %5d (%d Exact, %d Interpolated, %d Rounded).\n"),
 			MatchExact + MatchInter + MatchRound,
 			MatchExact, MatchInter, MatchRound);
-	printf("Failed:  %5d (%d Not matched, %d Write failure, %d Too Far,\n",
+	printf(_("Failed:  %5d (%d Not matched, %d Write failure, %d Too Far,\n"),
 			NotMatched + WriteFail + TooFar + NoDate + GPSPresent,
 			NotMatched, WriteFail, TooFar);
-	printf("                %d No Date, %d GPS Already Present.)\n",
+	printf(_("                %d No Date, %d GPS Already Present.)\n"),
 			NoDate, GPSPresent);
 
 
 	/* Clean up! */
-	FreePointList(Points);
-	if (GPSData) free(GPSData);
-	if (TimeAdjustment) free(TimeAdjustment);
-	if (Datum) free(Datum);
+	while (NumTracks > 0)
+	{
+		--NumTracks;
+		FreeTrack(&Track[NumTracks]);
+	}
+	free(Track);
+	free(Datum);
 	
-	return 0;
+	if (WriteFail)
+		/* A write failure is considered serious */
+		return EXIT_FAILURE;
+
+	/* Other failures aren't necessarily bad, depending on the input,
+	 * so provide a different return code to distinguish them.
+	 */
+	return(NotMatched + TooFar + NoDate + GPSPresent ? GPS_EXIT_WARNING : EXIT_SUCCESS);
 }
 
