@@ -6,7 +6,7 @@
  * the Unix time functions.
  */
 
-/* Copyright 2005 Daniel Foote.
+/* Copyright 2005-2012 Daniel Foote, Dan Fandrich.
  *
  * This file is part of gpscorrelate.
  *
@@ -25,9 +25,11 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+#include <sys/types.h>
+
 #define EXIF_DATE_FORMAT "%d:%d:%d %d:%d:%d"
 #define GPX_DATE_FORMAT "%d-%d-%dT%d:%d:%dZ"
 
-time_t ConvertToUnixTime(char* StringTime, char* Format,
+time_t ConvertToUnixTime(const char* StringTime, const char* Format,
 		int TZOffsetHours, int TZOffsetMinutes);
 
